'' =================================================================================================
''
''   File....... hc-8+_cauldron__161006.spin
''   Purpose.... Tim Young Cauldron Controller
''               -- 4 servos, 4 dimmers, AP-16+ control, PIR input
''   Author..... Jon "JonnyMac" McPhalen
''               Copyright (c) 2011-2016 Jon McPhalen
''               -- see below for terms of use
''   E-mail.....  
''   Started.... 
''   Updated.... 10 OCT 2016
''
'' =================================================================================================


con

  PCB_REV = "C"


con { timing }

  _clkmode = xtal1 + pll16x                                     
  _xinfreq = 5_000_000                                          ' use 5MHz crystal

  CLK_FREQ = (_clkmode >> 6) * _xinfreq                         ' system freq as a constant
  MS_001   = CLK_FREQ / 1_000                                   ' ticks in 1ms
  US_001   = CLK_FREQ / 1_000_000                               ' ticks in 1us

  EFX_BAUD =  34_800
  TRM_BAUD = 115_200


con { io pins }    

  RX1     = 31 { I }                                            ' programming / debug port
  TX1     = 30 { O }
  
  SDA     = 29 { IO }                                           ' eeprom
  SCL     = 28 { IO }

  OUT_EN  = 27 { O }                                            ' outputs enable pin (low)     
  
  SIO     = 26 { IO }                                           ' TTL serial io
  
  RX2     = 23 { I }                                            ' RS-485 port
  RX2_EN  = 22 { O }
  TX2_EN  = 21 { O }
  TX2     = 20 { O }

  R_LED   = 18 { O }                                            ' bi-color LED
  G_LED   = 17 { O }

  OUT7    = 15 { O }                                            ' outputs
  OUT6    = 14 { O }  
  OUT5    = 13 { O }
  OUT4    = 12 { O }
  OUT3    = 11 { O }
  OUT2    = 10 { O }  
  OUT1    =  9 { O }
  OUT0    =  8 { O }
               
  OPT_BR  =  7 { IO }                                           ' options swithes
  OPT_A1  =  6 { IO }  
  OPT_A0  =  5 { IO }
  OPT_SM  =  4 { IO }

  DMX_A8  =  3 { I }                                            ' high-bit off DMX address

  DO_165  =  2 { I }                                            ' control lines for 74x165
  CLK_165 =  1 { O }
  LD_165  =  0 { O }

  
  ' for EFX-TEK uSD adapter module (#99002)
  ' -- must open all option switches when installed

  SD_CS   =  OPT_BR                                       
  SD_DI   =  OPT_A1  
  SD_CLK  =  OPT_A0  
  SD_DO   =  OPT_SM


con { application }

  #0, OFF, RED, GRN, YEL                                        ' for r/g led
                                                                 
  #0, NOCON, RX, TX, RXTX                                       ' for rs-485
                                                                 
  #0, LSBFIRST, MSBFIRST                                        ' for 74x165 shift-register                                               
                                                                 
  #0, NO, YES                                                    
                                                                 
  NSERVOS  = 4                                                  ' number of servos
  NDIMMERS = 8 - NSERVOS                                        ' number of dimmers

  DEBUG    = YES
   

obj

' main                                                          ' * master Spin cog
  time   : "jm_time"                                            '   timing and delays
  prng   : "jm_prng"                                            '   pseudo-random number generator
  io     : "jm_io_basic"                                        '   essential io
  servos : "jm_servo8"                                          ' * servo driver     
  leds   : "jm_pwm8"                                            ' * pwm driver for dimmer outputs
  sd     : "fsrw"                                               ' * file IO
  audio  : "jm_fullduplexserial"                                ' * serial to AP-16+
  term   : "jm_fullduplexserial"                                ' * serial IO for terminal        
                                                                 
' * = consumes cog when running
  

var

  long  pirholdoff
  long  pirtimer
  
  ' for sd card
  
  word  channels                                                ' channels in Vixen file (4)
  word  eventms                                                 ' event timing; milliseconds
  word  frames                                                  ' frames (events) in project

  byte  buffer[8]                                               ' raw data buffer                                         
 

dat

  Version               byte    "?.?", 0                        ' for AP-16+
                         
  AmbientFile           byte    "vixen00.bin", 0
  AmbientAudio          byte    0                               ' play SFX00.WAV
                         
  MainFile              byte    "vixen01.bin", 0
  MainAudio             byte    1


pub main | t

  setup

  ' main loop
  ' -- ambient loop plays until PIR input detected
  ' -- animation loop plays one time

  repeat
    if (play_ambient == 1)                              ' pir input druing ambient?
      play_animation                                    '  yes, play animation
      pirholdoff := 20_000                              '  reset pir hold-off
      pirtimer   := 0


pub play_ambient | frametix, t, ch, pos, tix

'' Play background file
'' -- monitor PIR (on IN0, bit0 of digins)
'' -- returns 0 if no PIR, 1 if PIR detected

  ifnot (open_file(@AmbientFile, "r"))                  ' open file
    file_error(1)                                       ' oops, missing...

  ifnot (readbuf(@channels, 6))                         ' get header              
    file_error(2)                                       ' oops, bad header...     

    
  ' good file and valid header -- play
     
  if (DEBUG == YES)                                     ' show status
    term.str(string(13, "Playing: "))    
    term.str(@AmbientFile)           
    term.str(string(13, " -- "))
    term.dec(channels)
    term.str(string(" channels"))
    term.str(string(13, " -- "))
    term.dec(eventms)
    term.str(string("ms", 13))


  ap16_sfx(%00, AmbientAudio, 1)                        ' start audio  
    
  frametix := MS_001 * eventms                          ' ticks per frame
  set_rgled(OFF)
    
  t := cnt
  repeat
    ifnot (readbuf(@buffer, channels))                  ' read new frame
      quit                                              '  quit at end of file

    else
      repeat ch from 0 to NSERVOS-1                     ' update servos             
        pos := scale(buffer[ch], 255, 600, 2400)        ' convert to microseconds     
        servos.set(ch, pos, 0)                          ' update servo channel      
           
      if (channels > NSERVOS)                           ' update dimmers                                              
        repeat ch from NSERVOS to (channels-1)       
          leds.set(ch-NSERVOS, buffer[ch])

      if (++tix & %10)                                  ' show playing
        io.toggle(R_LED)
                                                        
      waitcnt(t += frametix)                            ' hold for frame time

      ' process pir  
          
      if (pirholdoff > 0)                               ' still in hold-off?         
        pirholdoff -= eventms                           ' yes, update
      else                                              ' no, scan pir
        if (ttl_pin == 0)                               ' IN0 active?
          pirtimer += eventms                           ' yes, update
          if (pirtimer => 100)                          ' valid?
            ap16_stop(%00)                              ' stop audio
            close_file                                  ' close ambient file
            set_rgled(OFF)
            return 1                                    ' return to main                                   
        else
          pirtimer := 0

  ap16_stop(%00)
  close_file
  set_rgled(OFF)
  return 0
      

pub play_animation | frametix, t, ch, pos, tix 
 
'' Play main animation file
'' -- run one time then return

  ifnot (open_file(@MainFile, "r"))                     ' open file
    file_error(3)                                       ' oops, missing...

  ifnot (readbuf(@channels, 6))                         ' get header              
    file_error(4)                                       ' oops, bad header...     

    
  ' good file and valid header -- play
     
  if (DEBUG == YES)                                     ' show status
    term.str(string(13, "Playing: "))    
    term.str(@MainFile)           
    term.str(string(13, " -- "))
    term.dec(channels)
    term.str(string(" channels"))
    term.str(string(13, " -- "))
    term.dec(eventms)
    term.str(string("ms", 13))


  ap16_sfx(%00, MainAudio, 1)                           ' start audio  

  frametix := MS_001 * eventms                          ' ticks per frame
  set_rgled(OFF)
    
  t := cnt
  repeat
    ifnot (readbuf(@buffer, channels))                  ' read new frame
      quit                                              '  quit at end of file

    else
      repeat ch from 0 to NSERVOS-1                     ' update servos             
        pos := scale(buffer[ch], 255, 600, 2400)        ' convert to microseconds     
        servos.set(ch, pos, 0)                          ' update servo channel      

      if (channels > NSERVOS)                           ' update dimmers                                              
        repeat ch from NSERVOS to (channels-1)       
          leds.set(ch-NSERVOS, buffer[ch])

      if (++tix & %10)                                  ' show playing
        io.toggle(G_LED)
          
      waitcnt(t += frametix)                            ' hold for frame time

      
  ap16_stop(%00)
  close_file
  set_rgled(OFF)
   
   
pub scale(raw, range, minout, maxout)

'' Scales raw (0 to range) value to new range: minout to maxout

  raw := 0 #> raw <# range                                      ' constrain input

  if (minout < maxout)                                          ' if good range
    return ((raw * (maxout - minout)) / range) + minout
  else
    return raw


pub file_error(mode) | idx

'' Flashes red LED if file error

  mode := 1 #> mode <# 5

  repeat
    repeat idx from 1 to mode
      set_rgled(RED)
      time.pause(100)
      set_rgled(OFF)
      time.pause(100)
    if (mode < 5)
      time.pause(1000 - (mode << 1))  


con

  { ----------------- }
  {  F I L E   I / O  } 
  { ----------------- }


var

  long  mounted
  

pub mount_sd | check

'' Attempt to mount SD card
'' -- returns true if SD card successfully mounted

  check := \sd.mount_explicit(SD_DO, SD_CLK, SD_DI, SD_CS)

  if (check => 0)
    mounted := true
  else
    mounted := false

  return mounted


pub dismount_sd

'' Unmounts SD card

  \sd.unmount
  mounted := false


pub open_file(spntr, mode) | check

'' Attempt to open file
'' -- spntr is pointer to filename (z-string)
'' -- mode is single character, (e.g., "r" for read)
'' -- sd card must be mounted before open

  if (mounted)
    check := \sd.popen(spntr, mode)                             ' attempt to open
    if (check => 0)
      if (mode == "r")
        \sd.seek(0)                                             ' force sector read 
      return true
    else
      return false
  else
    return false


pub close_file

'' Closes open file

  \sd.pclose


pub readln(spntr, n) | c, len

'' Reads line of text from open text file
'' -- terminated by LF or EOF
'' -- spntr is address of string buffer
'' -- n is maximum number of characters to read from line

  len := 0                                                     ' index into string

  repeat
    c := \sd.pgetc                                              ' get a character
    if (c < 0)                                                  ' end of file
      if (len == 0)                                             ' if nothing
        return -1                                               ' return empty
      else
        quit

    if (c <> 10)                                                ' if not EOL char
      byte[spntr++] := c                                        '  add to buffer
      len++
    else                                                        ' if LF
      quit                                                      '  we're done

  byte[spntr] := 0                                              ' terminate end of line

  return len
  

pub writeln(spntr, usecr, uself)

'' Writes z-string to open text file
'' -- set usecr to true to add CR
'' -- set uself to true to add LF

  \sd.pputs(spntr)
  if (usecr)
    \sd.pputc(13)
  if (uself)
    \sd.pputc(10)


pub readbuf(bpntr, n) | check

'' Reads n bytes from file to buffer at bpntr
'' -- returns true if good read

  check := \sd.pread(bpntr, n)
  if (check == n)
    return true
  else
    return false 


pub writebuf(bpntr, n) | check   

'' Writes n bytes to file from buffer at bpntr
'' -- returns true if good read

  check := \sd.pwrite(bpntr, n)
  if (check == n)
    return true
  else
    return false 


con

  { ------------------------------- }
  {  A P - 1 6 +   R O U T I N E S  } 
  { ------------------------------- }


pub ap16_setup(pin, baud)

'' Start audio serial driver on pin at baud (half-duplex true mode)

  audio.start(pin, pin, %1100, EFX_BAUD)
  time.pause(10)
  audio.rxflush


pub ap16_version(addr) | idx

'' Returns version string from AP-16+ at addr

  audio.str(string("!AP16"))
  audio.tx(addr)
  audio.tx("V")
  audio.rxflush

  repeat idx from 0 to 2
    byte[@Version][idx] := audio.rxtime(25)


pub ap16_status(addr)

'' Returns AP-16+ status byte

  audio.str(string("!AP16"))
  audio.tx(addr)
  audio.tx("G")
  audio.rxflush

  return audio.rxtime(25)


pub ap16_playing(addr) | status

'' Returns playing status from AP-16+ at addr

  return ((ap16_status(addr) & %1000_0000) > 0)
    
    
pub ap16_sfx(addr, sfx, rpts)

'' Plays SFX file on AP-16+ at addr

  audio.str(string("!AP16"))
  audio.tx(addr)
  audio.str(string("PS"))
  audio.tx(sfx)
  audio.tx(rpts)   


pub ap16_aux(addr, aux, rpts)

'' Plays AUX file on AP-16+ at addr

  audio.str(string("!AP16"))
  audio.tx(addr)
  audio.str(string("PA"))
  audio.tx(aux)
  audio.tx(rpts)


pub ap16_wav(addr, spntr, rpts)

'' Plays WAV file (name at spntr) on AP-16+ at addr

  audio.str(string("!AP16"))
  audio.tx(addr)
  audio.str(string("PW"))
  audio.str(spntr)
  audio.tx(13)      
  audio.tx(rpts)
   

pub ap16_stop(addr)

'' Stops AP-16+ at addr

  audio.str(string("!AP16"))
  audio.tx(addr)
  audio.tx("X")  

     
con

  { --------------------------------- }
  {  S U P P O R T   R O U T I N E S  } 
  { --------------------------------- }


var

  long  lastscan                                                ' ttl/DMX input scan


pub setup 

'' Configure for basic IO control

  outa[OUT7..OUT0] := %0000_0000                                ' clear outputs
  dira[OUT7..OUT0] := %1111_1111                                ' output mode                                  
  io.low(OUT_EN)                                                ' enable 74x245 

  io.high(LD_165)                                                  ' setup x165 io pins
  io.low(CLK_165)
  io.input(DO_165)

  set_rgled(OFF)
  
  set_rs485(NOCON)                                              ' RS-485 not connected

  ap16_setup(SIO, 38_400)                                       ' AP-16+ fast comms

  servos.start(NSERVOS, OUT0)                                   ' set servos

  if (NDIMMERS > 0)
    leds.start(NDIMMERS, OUT0 + NSERVOS)                        ' set leds

  if (DEBUG == YES)
    term.start(RX1, TX1, %0000, TRM_BAUD)
  
  time.pause(3_000)                                             ' let AP-16+ get ready

  if (DEBUG == YES)
    term.str(string(term#CLS, "Starting program...", 13))

  ' Look for uSD card on HC-8+
  ' -- make sure all OPTION switches are set to OFF
  ' -- waits for uSD insertion
  ' -- flash at 1Hz if no card found

  repeat
    if (mount_sd == false)
      set_rgled(RED)
      time.pause(500)
      set_rgled(OFF)
      time.pause(500)  
        
    else
      set_rgled(OFF)
      quit

  pirholdoff := 20_000                                  ' wait at least 20s
  pirtimer   := 0    


pub ttl_pin | btns

'' Reads TTL header for 1-of-8 input
'' -- returns -1 (no input)
'' -- returns 0 to 7 for input on IN0 to IN7
''    * only one input allowed

  btns := ttl_inputs(true)                                      ' read ttl inputs

  if (bit_count(btns) == 1)                                     ' only one button pressed?
    return bit_pos(btns, 0)                                     '  yes, return button #
  else
    return -1                                                   '  none or bad input


pub ttl_inputs(rescan) 

'' Return TTL inputs status
'' -- updates global var "lastscan" if rescan is true

  if (rescan)
    lastscan := x165_in(16, MSBFIRST) | (ina[DMX_A8] << 16)     ' read in0..in7, dmx addresss
   
  return (lastscan & $FF)                                       ' return TTL input bits


pub dmx_address(rescan) 

'' Return DMX address switch setting
'' -- updates global var "lastscan" if rescan is true

  if (rescan)
    lastscan := x165_in(16, MSBFIRST) | (ina[DMX_A8] << 16)     ' read in0..in7, dmx addresss

  return (lastscan >> 8) & $01FF                                ' return DMX address bits
 

pub x165_in(bits, mode) | tmp165

'' Input value from 74x165(s)      
                                                                                          
  outa[LD_165] := 0                                             ' blip Shift/Load line    
  outa[LD_165] := 1                                                                         
                                                                                          
  tmp165 := 0                                                   ' clear workspace
  bits := 1 #> bits <# 32                                       ' limit to legal value         
  repeat bits                                                   ' read n bits         
    tmp165 := (tmp165 << 1) | ina[DO_165]                       ' get new bit             
    outa[CLK_165] := 1                                          ' blip clock              
    outa[CLK_165] := 0  

  if (mode == LSBFIRST)                                         ' LSBFIRST result?
    tmp165 ><= bits                                             ' reverse bits
   
  return tmp165
  

pub option_inputs

'' Returns state of option switch bits (%0000..%1111, all off to all on)

  dira[OPT_BR..OPT_SM] := %0000                                 ' force to input mode

  return !ina[OPT_BR..OPT_SM]                                   ' return state (1 = on)


pub set_rgled(state)

'' Sets R/G LED
'' -- direct control only

  case state
    OFF : outa[R_LED..G_LED] := %00
    RED : outa[R_LED..G_LED] := %10
    GRN : outa[R_LED..G_LED] := %01
    YEL : outa[R_LED..G_LED] := %00                             ' yellow requires PASM driver

  dira[R_LED..G_LED] := %11
      

pub set_rs485(state)

'' Sets RS-485 RX and TX enable inputs

  case state
    NOCON : outa[RX2_EN..TX2_EN] := %10
    RX    : outa[RX2_EN..TX2_EN] := %00 
    TX    : outa[RX2_EN..TX2_EN] := %11   
    RXTX  : outa[RX2_EN..TX2_EN] := %01

  dira[RX2_EN..TX2_EN] := %11


pub bit_count(value) | bc

'' Returns # of bits set (1) in value

  bc := 0                                                       ' clear count
  repeat 32                                                     ' test all bits
    bc += (value & %1)                                          ' add bit value to count
    value >>= 1                                                 ' next bit

  return bc


pub bit_pos(value, mode)

'' Returns position of 1st bit set (0..31)
'' -- mode 0 to scan from lsb, mode 1 to scan from msb 
'' -- -1 = no bits set

  if (value == 0)                                               ' if no bits                                   
    return -1                                                   '  return -1

  else
    if (mode == 0)                                              ' check from LSB
      value ><= 32                                              '  flip for >|
      return (32 - >|value)
    else
      return (>|value - 1)


pub bit_val(value, pos)

'' Returns bit (0..1) from pos (0..31) in value 

  if ((pos => 0) and (pos =< 31))
    return (value >> pos) & 1
  else
    return 0

    
dat

{{

  Copyright (c) 2011-2016 Jon McPhalen 

  Terms of Use: MIT License

  Permission is hereby granted, free of charge, to any person obtaining a copy of this
  software and associated documentation files (the "Software"), to deal in the Software
  without restriction, including without limitation the rights to use, copy, modify,
  merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
  permit persons to whom the Software is furnished to do so, subject to the following
  conditions:

  The above copyright notice and this permission notice shall be included in all copies
  or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
  INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
  PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
  CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
  OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 

}}      